/*
  scicalc version 1.0
  This software is public domain. It may not be resold.

  Written by Mark Overmars
  Date 26 Nov 1990

  structure generalised and scientific functions added by 
  Robin Humble, 5-2-91, (big bug hunter: James Murray)

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "forms.h"

#define PI      3.14159265358979323846

/* operators: */
#define PLUS    0
#define MIN     1
#define MUL     2
#define DIV     3
#define POW     4
#define INVPOW  5
#define IS      6

/* display modes: */
#define DOT    -1
#define NORM   -2
#define EXP    -3

#define MaxNumBrkts 10

int priority[6] = {1, 1, 2, 2, 3, 3};   /* the priorities of each operator */

double value[4*(MaxNumBrkts+1)];    /* The values on the stack */
int oper[3*(MaxNumBrkts+1)];        /* the operators between them */
int top;                            /* the top of the stack */

int startbrkt[(MaxNumBrkts+1)];     /* the positions of the left brackets */
int currentbrkt;                    /* bracketing we are in */

double mem;             /* The memory value */

int ready;              /* Whether last number is ready.
                           if "ready" is set, then typing another number
                           overwrites the current number. */
int dot;                /* Whether the dot has been typed */
double diver;           /* The divider when behind the dot */
int behind;             /* Number of digits behind dot */

int inv;                /* Whether inverse key is depressed */

int emode;              /* Whether we are entering the exponent */
int exponent;           /* the exponent value whilst entering exponent */
double mantissa;        /* the mantissa value whilst entering exponent */

int base;               /* the base we are working in (2,8,10 or 16) */

int drgmode;            /* whether we are in deg, rad or grad mode */

#include "calcform.c"

double mytrunc(double x)   /* not implemented on my system for some reason */
{
  if (x < 0.0)
    x = -floor(-x);
  else
    x = floor(x);

  return x;
}

void setnormlabels()
{
  fl_freeze_form(form);
  if (base <= 10) {
    fl_set_object_label(sqrtbut, "sqrt");
    fl_set_object_label( powbut, "x^y");
    fl_set_object_label( sinbut, "sin");
    fl_set_object_label( cosbut, "cos");
    fl_set_object_label( tanbut, "tan");
    fl_set_object_label( logbut, "log");
    fl_set_object_lcol(sqrtbut,4);
    fl_set_object_lcol( powbut,4);
    fl_set_object_lcol( sinbut,4);
    fl_set_object_lcol( cosbut,4);
    fl_set_object_lcol( tanbut,4);
    fl_set_object_lcol( logbut,4);
  }
  else {
    fl_set_object_label(sqrtbut, "a");
    fl_set_object_label( powbut, "b");
    fl_set_object_label( sinbut, "c");
    fl_set_object_label( cosbut, "d");
    fl_set_object_label( tanbut, "e");
    fl_set_object_label( logbut, "f");
    fl_set_object_lcol(sqrtbut,0);
    fl_set_object_lcol( powbut,0);
    fl_set_object_lcol( sinbut,0);
    fl_set_object_lcol( cosbut,0);
    fl_set_object_lcol( tanbut,0);
    fl_set_object_lcol( logbut,0);
  }    
  fl_set_object_label(    lnbut, "ln");
  fl_set_object_label(   intbut, "int");
  fl_set_object_label(  dtorbut, "d->r");
  fl_set_object_label(memaddbut, "M+");
  fl_set_object_label(memmulbut, "M*");
  fl_set_object_label(  memcbut, "MC");
  fl_unfreeze_form(form);
}

void setinvlabels()
{
  fl_freeze_form(form);
  if (base <= 10) {
    fl_set_object_label(  sqrtbut, "x^2");
    fl_set_object_label(   powbut, "x^1/y");
    fl_set_object_label(   sinbut, "arsin");
    fl_set_object_label(   cosbut, "arcos");
    fl_set_object_label(   tanbut, "artan");
    fl_set_object_label(   logbut, "10^x");
    fl_set_object_lcol(sqrtbut,4);
    fl_set_object_lcol( powbut,4);
    fl_set_object_lcol( sinbut,4);
    fl_set_object_lcol( cosbut,4);
    fl_set_object_lcol( tanbut,4);
    fl_set_object_lcol( logbut,4);
  }
  else {
    fl_set_object_label(  sqrtbut, "a");
    fl_set_object_label(   powbut, "b");
    fl_set_object_label(   sinbut, "c");
    fl_set_object_label(   cosbut, "d");
    fl_set_object_label(   tanbut, "e");
    fl_set_object_label(   logbut, "f");
    fl_set_object_lcol(sqrtbut,0);
    fl_set_object_lcol( powbut,0);
    fl_set_object_lcol( sinbut,0);
    fl_set_object_lcol( cosbut,0);
    fl_set_object_lcol( tanbut,0);
    fl_set_object_lcol( logbut,0);
  }    
  fl_set_object_label(    lnbut, "e^x");
  fl_set_object_label(   intbut, "frac");
  fl_set_object_label(  dtorbut, "r->d");
  fl_set_object_label(memaddbut, "M-");
  fl_set_object_label(memmulbut, "M/");
  fl_set_object_label(  memcbut, "Mex");
  fl_unfreeze_form(form);
}

void cvttobase(double num, int base, int behind, char *str)
{
  double sign, div;
  int place, digit, i;
  char digstr[2];

  sign = copysign(1.0, num);
  num *= sign;
  if (sign == -1.0)
    sprintf(str, "-");
  else
    sprintf(str, "");

  if (num == 0.0) {
    sprintf(str, "0");
    if (behind > 0) {
      strcat(str, ".");
      for(i = 0; i < behind; i++) strcat(str, "0");
    }
    return;
  }

  place = (int)( log(num)/log((double)base) );
  if (place < 0) place = 0;

  do {
    div = pow((double)base, (double)place);
    digit = (int)(num/div);
    num -= (double)digit*div;
    if (place == -1) strcat(str, ".");
    place--;
    sprintf(digstr, "%x", digit);
    strcat(str, digstr);
    if (strlen(str) > 18) {
      sprintf(str, "can't display");
      return;
    }
  } while ((place >= 0) || ((place >= -9) && (num != 0.0)));

  if ((place == -1) && ((behind == DOT) || (behind > 0))) strcat(str, ".");
  while ((behind > 0) && (behind >= -place)) {
    strcat(str, "0");
    place--;
  }
}

void set_display(double val, int behind)
{
  int i;
  char dispstr[40], expstr[10], str2[10];

  /* number or operator handled to get here so reset inv stuff */

  if (inv) {
    inv = 0;
    fl_set_button(invbut, 0);
    setnormlabels();
  }

  if (behind >= 0) { /* format with appropriate number of decimal places */
    if (base == 10) {
      emode = 0;
      strcpy(str2,"%.1f");
      str2[2] = behind + '0';
      sprintf(dispstr,str2,val);
    }
    else /* non base 10 display */
      cvttobase(val, base, behind, dispstr);
  }
  else if (behind == DOT) { /* display the . at the right */
    if (base == 10) {
      emode = 0;
      sprintf(dispstr,"%.1f",val);
      dispstr[strlen(dispstr)-1] = NULL;
    }
    else
      cvttobase(val, base, behind, dispstr);
  }
  else if (behind == NORM) { /* normal display */
    if (base == 10) {
      emode = 0;
      sprintf(dispstr,"%.9g",val);
    }
    else /* non base 10 display */
      cvttobase(val, base, behind, dispstr);
  }
  else { /* exponent entering display */
    sprintf(dispstr,"%.8f",val);
    for (i = strlen(dispstr); dispstr[i-1] == '0'; i--);
    dispstr[i] = NULL;
    strcat(dispstr, "e");
    sprintf(expstr,"%d",exponent);
    strcat(dispstr, expstr);
  }

  strcat(dispstr," ");
  dispstr[17] = NULL;
  fl_set_object_label(display,dispstr);
}

void set_memdisp()
{
  if (mem)
    fl_set_object_label(memdisp,"M");
  else
    fl_set_object_label(memdisp,"");
}

void set_drgdisp()
{
  if (drgmode == 0)
    fl_set_object_label(drgdisp, "DEG");
  else if (drgmode == 1)
    fl_set_object_label(drgdisp, "RAD");
  else
    fl_set_object_label(drgdisp, "GRAD");
}

void set_brktdisp()
{
  char dispstr[40];

  if (currentbrkt > 0) {
    sprintf(dispstr, "%d [ max %d", currentbrkt, MaxNumBrkts);
    fl_set_object_label(brktdisp, dispstr);
  }
  else
    fl_set_object_label(brktdisp,"");
}

void init_value(int lev)
/* initializes adding a value to the stack */
{
  top = lev;
  value[top] = 0.0;
  ready = 0;
  emode = 0;
  dot = 0;
  diver = 1.0;
  behind = 0;

  if (inv) {
    inv = 0;
    fl_set_button(invbut, 0);
    setnormlabels();
  }
}

void handle_numb(double numb)
/* handles typing a number */
{
  int first;
  double sign;

  if (ready) init_value(top);

  if (numb == -1.0) 
    if (dot) /* check whether we already have a dot */
      return;
    else {
      dot = 1;
      set_display(value[top],DOT);
      return;
    }

  if (emode) {
    sign = copysign(1.0, (double)exponent);
    if (abs(exponent)*10 + numb > 999) { /* cycle if exponent has > 3 digits */
      first = (int)floor((double)abs(exponent)/100.0);
      exponent = abs(exponent) - 100*first;
      exponent *= (int)sign;
    }
    exponent = exponent*10 + (int) (sign*numb);
    value[top] = mantissa*pow(10.0, (double)exponent);
    set_display(mantissa, EXP);
  }
  else if (numb < base) { /* both decimal and non decimal number entry */
    sign = copysign(1.0, value[top]);
    if (dot && behind < 9) {
      behind++;
      diver = diver/(double)base;
      value[top] += sign*diver*numb;
    }
    else if ((! dot) && (value[top] < 1.0e10))
        value[top] = (double)base*value[top] + sign*numb;

    set_display(value[top],behind);
  }
}

void calc(int i)
/* calculates an entry in the stack */
{
  switch (oper[i])
  {
    case PLUS:   value[i] += value[i+1]; break;
    case MIN:    value[i] -= value[i+1]; break;
    case MUL:    value[i] *= value[i+1]; break;
    case DIV:    value[i] /= value[i+1]; break;
    case POW:    value[i] = pow(value[i], value[i+1]); break;
    case INVPOW: value[i] = pow(value[i], 1.0/value[i+1]); break;
  }
}

void changebase(int newbase)
{
  int oldbase;

  oldbase = base;
  base = newbase;

  set_display(value[top], NORM);
  ready = 1;
  if ((oldbase == 16) || (base == 16)) setnormlabels();
}

void exch()
/* handles an exchange of the last 2 values */
{
  double temp;

  /* check if we have 2 values to exchange */
  if (top > startbrkt[currentbrkt]) {
    temp = value[top];
    value[top] = value[top-1];
    value[top-1] = temp;

    set_display(value[top],NORM);
    ready = 1;
  }
}


void memexch()
/* handles an exchange between memory and last val */
{
  double temp;

  temp = mem;
  mem = value[top];
  value[top] = temp;

  set_display(value[top],NORM);
  ready = 1;
  set_memdisp();
}


void factorial()
/* handles a factorial */
{
  extern int signgam;
  double lg, alpha;

  /* uses gamma functions to get result for non-integer values */

  alpha = value[top] + 1.0;
  if ((floor(alpha) == alpha)&&(alpha <= 0.0)) {
    init_value(0);
    fl_set_object_label(display,"Error: -ve integer ");
  }
  else {
    lg = gamma(alpha);
    value[top] = (double)signgam*exp(lg);
    set_display(value[top],NORM);
    ready = 1;
  }
}

void Epi()
/* handles exponent/pi button */
{
  if ((value[top] == 0.0) || (ready)) {
    value[top] = PI;
    set_display(value[top],NORM);
    ready = 1;
  }
  else if ((! emode) && (base == 10)) {
    emode = 1;
    exponent = 0;
    mantissa = value[top];
    set_display(mantissa,EXP);
  }
}

double todrg(double angle)
{
  if (drgmode == 0)
    return (PI*angle/180.0);
  else if (drgmode == 2)
    return (PI*angle/100.0);
  else
    return (angle);
}

double fromdrg(double angle)
{
  if (drgmode == 0)
    return (180.0*angle/PI);
  else if (drgmode == 2)
    return (100.0*angle/PI);
  else
    return (angle);
}

void addlbrkt()
/* handles a left bracket */
{
  if (currentbrkt < MaxNumBrkts) {
    currentbrkt++;
    startbrkt[currentbrkt] = top;
    ready = 1;
    set_brktdisp();
  }
}

void addrbrkt()
/* handles a right bracket */
{
  int i;

  if (currentbrkt > 0) {
    for (i = top; i > startbrkt[currentbrkt]; i--) calc(i-1);
    top = startbrkt[currentbrkt];
    currentbrkt--;
    ready = 1;
  }

  set_display(value[top],NORM);
  set_brktdisp();
}

void handle_oper(int op)
/* handles an operator */
{
  int prevop, i, finished;

  switch (op)
  {
    case PLUS:
    case MIN:
    case MUL:
    case DIV:
    case POW:
    case INVPOW:
      finished = 0;
      do {
        if (top == startbrkt[currentbrkt]) finished = 1; /* is 1st operator */
        if (! finished) { /* compare priority of previous
                             operators with current op */
          prevop = oper[top-1];
          if (priority[prevop] < priority[op])
            finished = 1;
          else { /* last op can be calculated */
            top--;
            calc(top);
          }
        }
      } while (! finished);

      oper[top] = op;
      init_value(top+1);

      set_display(value[top-1],NORM);
      break;

    case IS:
      while (currentbrkt > 0) addrbrkt();
      for (i = top; i > 0; i--) calc(i-1);
      top = 0;
      ready = 1;
      set_display(value[top],NORM);
      break;
  }
}

void handle_but(FL_OBJECT *but)
/* handles a button pressed by the user */
{
  /* number buttons: */
  if      (but == onebut)   handle_numb(1.0);
  else if (but == twobut)   handle_numb(2.0);
  else if (but == threebut) handle_numb(3.0);
  else if (but == fourbut)  handle_numb(4.0);
  else if (but == fivebut)  handle_numb(5.0);
  else if (but == sixbut)   handle_numb(6.0);
  else if (but == sevenbut) handle_numb(7.0);
  else if (but == eightbut) handle_numb(8.0);
  else if (but == ninebut)  handle_numb(9.0);
  else if (but == zerobut)  handle_numb(0.0);

  else if (but == dotbut)   handle_numb(-1.0);

  /* operator buttons: */
  else if (but == plusbut)  handle_oper(PLUS);
  else if (but == minbut)   handle_oper(MIN);
  else if (but == mulbut)   handle_oper(MUL);
  else if (but == divbut)   handle_oper(DIV);
  else if (but == powbut)
    if (base > 10)
      handle_numb(11.0);
    else
      if (! inv) handle_oper(POW); else handle_oper(INVPOW);

  else if (but == isbut)    handle_oper(IS);

  /* bracket buttons: */
  else if (but == lbrktbut) addlbrkt();
  else if (but == rbrktbut) addrbrkt();

  /* misc buttons: */
  else if (but == Epibut)   Epi();
  else if (but == factbut)  factorial();
  else if (but == exchbut)  exch();
  else if (but == pmbut)
    if (! emode) {
      value[top] = -value[top];
      set_display(value[top],NORM);}
    else {exponent = -exponent;
      value[top] = mantissa*pow(10.0, (double)exponent);
      set_display(mantissa,EXP);
  }
  else if (but == recipbut) {
    value[top] = 1.0/value[top];
    set_display(value[top],NORM);
    ready = 1;
  }

  /* clear buttons: */
  else if (but == clearbut) {init_value(top); set_display(0.0,NORM);}
  else if (but == allcbut)  {
    init_value(0);
    set_display(0.0,NORM);
    currentbrkt = 0;
    fl_set_object_label(brktdisp,"");
  }

  /* memory buttons: */
  else if (but == memcbut) 
    if (! inv){
      mem = 0.0;
      set_display(value[top],NORM);
      ready = 1;
      set_memdisp();
    }
    else memexch();
  else if (but == memrclbut) {
    value[top] = mem;
    set_display(value[top],NORM);
    ready = 1;
  }
  else if (but == memstobut) {
    mem = value[top];
    set_display(value[top],NORM);
    ready = 1;
    set_memdisp();
  }
  else if (but == memaddbut) {
    if (! inv) mem += value[top]; else mem -= value[top];
    set_display(value[top],NORM);
    ready = 1;
    set_memdisp();
  }
  else if (but == memmulbut) {
    if (! inv) mem *= value[top]; else mem /= value[top];
    set_display(value[top],NORM);
    ready = 1;
    set_memdisp();
  }

  /* deg-rad-grad button: */
  else if (but == drgbut) {drgmode++; drgmode %= 3; set_drgdisp();}

  /* inverse button: */
  else if (but == invbut)
    if (inv) {inv = 0; setnormlabels();}
    else {inv = 1; setinvlabels();}

  /* change base buttons: */
  else if (but == base2but)  changebase(2);
  else if (but == base8but)  changebase(8);
  else if (but == base10but) changebase(10);
  else if (but == base16but) changebase(16);

  /* buttons with inverses: */
  else if (but == sqrtbut)
    if (base > 10) handle_numb(10.0); else
    if (! inv) {value[top] = sqrt(value[top]); set_display(value[top],NORM); ready = 1;}
    else       {value[top] = pow(value[top], 2.0); set_display(value[top],NORM); ready = 1;}
  else if (but == sinbut)
    if (base > 10) handle_numb(12.0); else
    if (! inv) {value[top] = sin(todrg(value[top])); set_display(value[top],NORM); ready = 1;}
    else       {value[top] = fromdrg(asin(value[top])); set_display(value[top],NORM); ready = 1;}
  else if (but == cosbut)
    if (base > 10) handle_numb(13.0); else
    if (! inv) {value[top] = cos(todrg(value[top])); set_display(value[top],NORM); ready = 1;}
    else       {value[top] = fromdrg(acos(value[top])); set_display(value[top],NORM); ready = 1;}
  else if (but == tanbut)
    if (base > 10) handle_numb(14.0); else
    if (! inv) {value[top] = tan(todrg(value[top])); set_display(value[top],NORM); ready = 1;}
    else       {value[top] = fromdrg(atan(value[top])); set_display(value[top],NORM); ready = 1;}
  else if (but == logbut)
    if (base > 10) handle_numb(15.0); else
    if (! inv) {value[top] = log10(value[top]); set_display(value[top],NORM); ready = 1;}
    else       {value[top] = pow(10.0, value[top]); set_display(value[top],NORM); ready = 1;}
  else if (but == lnbut)
    if (! inv) {value[top] = log(value[top]); set_display(value[top],NORM); ready = 1;}
    else       {value[top] = exp(value[top]); set_display(value[top],NORM); ready = 1;}
  else if (but == intbut)
    if (! inv) {value[top] = mytrunc(value[top]); set_display(value[top],NORM); ready = 1;}
    else       {value[top] = value[top] - mytrunc(value[top]); set_display(value[top],NORM); ready = 1;}
  else if (but == dtorbut)
    if (! inv) {value[top] = PI*value[top]/180.0; set_display(value[top],NORM); ready = 1;}
    else       {value[top] = 180.0*value[top]/PI; set_display(value[top],NORM); ready = 1;}

}


main()
{
  FL_OBJECT *obj;
  create_the_forms();

  init_value(0);
  drgmode = 1;
  base = 10;
  currentbrkt = 0;
  startbrkt[0] = 0;

  set_memdisp();
  set_brktdisp();
  fl_set_button(base10but, 1);
  set_drgdisp();
  set_display(0.0,0);

  fl_show_form(form,FL_PLACE_SIZE,FALSE,NULL);
  do
  {
    obj = fl_do_forms();
    handle_but(obj);
  } while (obj != exitbut);
  fl_hide_form(form);
}
